/*************************************************************************\
 * Copyright (c) 2022 The University of Chicago, as Operator of Argonne
 * National Laboratory.
 * Copyright (c) 2022 The Regents of the University of California, as
 * Operator of Los Alamos National Laboratory.
 * This file is distributed subject to a Software License Agreement found
 * in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#include "mdb.h"
#include "SDDS.h"
#include "scan.h"

#define CLO_FINDLOCATIONOF 0
#define CLO_GRIDVARIABLES 1
#define CLO_VALUESFILE 2
#define CLO_ATVALUES 3
#define CLO_PIPE 4
#define CLO_INTERPOLATE 5
#define CLO_MODE 6
#define CLO_PRESORTED 7
#define CLO_OPTIONS 8

char *option[CLO_OPTIONS] = {
  "findlocationof",
  "gridvariables",
  "valuesfile",
  "atvalues",
  "pipe",
  "interpolate",
  "mode",
  "presorted",
};

char *USAGE = 
"sddsfindin2dgrid [<input>] [<output>] [-pipe=[input][,output]]\n\
-gridVariables=<gridColumnName1>,<gridColumnName2> [-presorted]\n\
-findLocationOf=<columnName1>,<columnName2> {-valuesFile=<filename> | -atValues=<value1>,<value2>]}\n\
[-interpolate] [-mode={onePairPerPage|reuseFirstPage|all}]\n\
The input file contains a grid of <gridColumnName1> and <gridColumnName2>, along with \n\
<columnName1> and <columnName2>. This program searches the grid to find the location\n\
(<gridColumnName1>,<gridColumnName2>) where <columnName1> and <columnName2> are closest to given values.\n\
-gridVariables  Names the two columns that are laid out on a grid.\n\
-presorted      Data is sorted by <gridColumnName1> and <gridColumnName2>, e.g., using\n\
                sddssort -column=<gridColumnName1> -column=<gridColumn2>.\n\
                Pre-sorting can save considerable time if data is used repeatedly.\n\
-findLocationOf Names the two columns that are to be located on the grid by finding\n\
                the optimal values of the gridVariables.\n\
-valuesFile     Names a file from which a series of pairs (<columnName1>,<columnName2>) will be\n\
                taken.\n\
-atValues       Gives values of (<columnName1>,<columnName2>) to be found. May be repeated.\n\
-interpolate    By default, the location is the optimal point on the grid. If this option is \n\
                given, 2d linear interpolation is performed to refine the location.\n\
-mode           By default, one successive pair of values is used with each successive page from <input>;\n\
                this is -mode=onePairPerPage. -mode=reuseFirstPage means that all values are used\n\
                with the first page of <input>; other pages are ignored. -mode=all means that\n\
                all values are used with all pages.\n\n\
Program by Michael Borland. (" __DATE__ " " __TIME__ ", SVN revision: " SVN_VERSION ")\n";

#define MODE_ONEPAIRPERPAGE 0x01UL
#define MODE_REUSEFIRSTPAGE 0x02UL
#define MODE_ALL 0x04UL 

static double *gridValue[2]={NULL,NULL}, *valueAtLocation[2]={NULL,NULL}; 
static uint64_t ng[2]={0,0};

void gridifyData(char *gridVariable[2], uint64_t gridPoints, short presorted);
int findLocationInGrid(double at1, double at2,  double *location, double *value, short interpolate);

int main(int argc, char **argv)
{
  long iArg;
  SDDS_DATASET SDDSin, SDDSout, SDDSvalues;
  SCANNED_ARG *scanned;
  unsigned long pipeFlags, flags;
  uint64_t gridPoints, atValues, iv, irow;
  char *input, *output, *fileForValues;
  char *findLocationOf[2], *gridVariable[2];
  double *atValue[2], value[2];
  short interpolate, restarted, needPage, presorted;
  unsigned long mode;

  SDDS_RegisterProgramName(argv[0]);

  argc = scanargs(&scanned, argc, argv);
  if (argc == 1)
    bomb(NULL, USAGE);

  input = output = fileForValues = NULL;
  pipeFlags = flags = 0;
  atValues = 0;
  atValue[0] = atValue[1] = NULL;
  findLocationOf[0] = findLocationOf[1] = NULL;
  gridVariable[0] = gridVariable[1] = NULL;
  interpolate = presorted = 0;
  mode = MODE_ALL;

  for (iArg = 1; iArg < argc; iArg++) {
    if (scanned[iArg].arg_type == OPTION) {
      /* process options here */
      switch (match_string(scanned[iArg].list[0], option, CLO_OPTIONS, 0)) {
      case CLO_FINDLOCATIONOF:
        if (scanned[iArg].n_items != 3 ||
            !strlen(findLocationOf[0] = scanned[iArg].list[1]) ||
            !strlen(findLocationOf[1] = scanned[iArg].list[2]))
          SDDS_Bomb("invalid -findLocationOf syntax");
        if (strcmp(findLocationOf[0], findLocationOf[1])==0)
          SDDS_Bomb("invalid -findLocationOf values---two variables are the same");
        break;
      case CLO_GRIDVARIABLES:
        if (scanned[iArg].n_items != 3 ||
            !strlen(gridVariable[0] = scanned[iArg].list[1]) ||
            !strlen(gridVariable[1] = scanned[iArg].list[2]))
          SDDS_Bomb("invalid -gridVariables syntax");
        if (strcmp(gridVariable[0], gridVariable[1])==0)
          SDDS_Bomb("invalid -gridVariables values---two variables are the same");
        break;
      case CLO_VALUESFILE:
        if (scanned[iArg].n_items!=2 ||
            !strlen(fileForValues=scanned[iArg].list[1]))
          SDDS_Bomb("invalid -valuesFile syntax");
        if (atValues>0)
          SDDS_Bomb("can't give -valuesFile and -atValues together");
        break;
      case CLO_ATVALUES:
        atValue[0] = SDDS_Realloc(atValue[0], sizeof(double)*(atValues+1));
        atValue[1] = SDDS_Realloc(atValue[1], sizeof(double)*(atValues+1));
        if (scanned[iArg].n_items!=3 ||
            sscanf(scanned[iArg].list[1], "%le", &atValue[0][atValues])!=1 ||
            sscanf(scanned[iArg].list[2], "%le", &atValue[1][atValues])!=1 )
          SDDS_Bomb("invalid -atValues syntax");
        if (fileForValues)
          SDDS_Bomb("can't give -valuesFile and -atValues together");
        atValues++;
        break;
      case CLO_PIPE:
        if (!processPipeOption(scanned[iArg].list + 1, scanned[iArg].n_items - 1, &pipeFlags))
          SDDS_Bomb("invalid -pipe syntax");
        break;
      case CLO_INTERPOLATE:
        interpolate = 1;
        break;
      case CLO_PRESORTED:
        presorted = 1;
        break;
      case CLO_MODE:
        mode = 0;
        if ((scanned[iArg].n_items-=1)!=1 ||
            !scanItemList(&mode, scanned[iArg].list+1, &scanned[iArg].n_items, 0,
                          "onepairperpage", -1, NULL, 0, MODE_ONEPAIRPERPAGE,
                          "reusefirstpage", -1, NULL, 0, MODE_REUSEFIRSTPAGE, 
                          "all", -1, NULL, 0, MODE_ALL,
                          NULL) ||
            bitsSet(mode)!=1)
          SDDS_Bomb("invalid -mode syntax");
        break;
      default:
        fprintf(stderr, "invalid option seen: %s\n", scanned[iArg].list[0]);
        exit(1);
        break;
      }
    } else {
      if (!input)
        input = scanned[iArg].list[0];
      else if (!output)
        output = scanned[iArg].list[0];
      else
        SDDS_Bomb("too many filenames");
    }
  }

  if (!findLocationOf[0] || !findLocationOf[1])
    SDDS_Bomb("must give -findLocationOf");
  if (!gridVariable[0] || !gridVariable[1])
    SDDS_Bomb("must give -gridVariables");
  if (!atValues && !fileForValues)
    SDDS_Bomb("must give -atValues or -fileForValues");

  processFilenames("sddsfindin2dgrid", &input, &output, pipeFlags, 0, NULL);

  if (fileForValues) {
    if (!SDDS_InitializeInput(&SDDSvalues, fileForValues))
      SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
    if (SDDS_ReadPage(&SDDSvalues)<=0)
      SDDS_Bomb("unable to read values file");
    if ((atValues = SDDS_RowCount(&SDDSvalues))>0) {
      if (!(atValue[0] = SDDS_GetColumnInDoubles(&SDDSvalues, findLocationOf[0])))
        SDDS_Bomb("unable to retrieve values of first \"findLocationOf\" variable");
      if (!(atValue[1] = SDDS_GetColumnInDoubles(&SDDSvalues, findLocationOf[1])))
        SDDS_Bomb("unable to retrieve values of second \"findLocationOf\" variable");
    }
    if (SDDS_ReadPage(&SDDSvalues)>0)
      SDDS_Bomb("values file has multiple pages, which is not supported");
    SDDS_Terminate(&SDDSvalues);
    SDDS_ClearErrors();
  }

  if (!SDDS_InitializeInput(&SDDSin, input))
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);

  if (!SDDS_InitializeOutput(&SDDSout, SDDS_BINARY, 0, NULL, NULL, output) ||
      !SDDS_TransferAllParameterDefinitions(&SDDSout, &SDDSin, 0) ||
      !SDDS_TransferColumnDefinition(&SDDSout, &SDDSin, gridVariable[0], NULL) ||
      !SDDS_TransferColumnDefinition(&SDDSout, &SDDSin, gridVariable[1], NULL) ||
      !SDDS_TransferColumnDefinition(&SDDSout, &SDDSin, findLocationOf[0], NULL) ||
      !SDDS_TransferColumnDefinition(&SDDSout, &SDDSin, findLocationOf[1], NULL) ||
      !SDDS_WriteLayout(&SDDSout) ||
      !SDDS_StartPage(&SDDSout, atValues*1000))
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
  SDDS_ClearErrors();

  irow = 0;
  iv = 0;
  restarted = 0;
  while (1) {
    double location[2];
    needPage = irow==0 ? 1 : 0;
    if (mode==MODE_ONEPAIRPERPAGE) {
      if (iv==atValues)
        break;
      needPage = 1;
    } else if (mode==MODE_REUSEFIRSTPAGE) {
      if (iv==atValues)
        break;
      if (iv==0)
        needPage = 1;
    } else if (mode==MODE_ALL) {
      if (iv==atValues) {
        needPage = 1;
        iv = 0;
        restarted = 1; /* signals that SDDS_ReadPage<=0 is ok */
      }
    }

    if (needPage) {
      if (SDDS_ReadPage(&SDDSin)<=0) {
        if (!restarted)
          SDDS_Bomb("too few pages in input file for number of location requests");
        break;
      }
      if (gridValue[0]) {
        free(gridValue[0]);
        free(gridValue[1]);
        free(valueAtLocation[0]);
        free(valueAtLocation[1]);
        gridValue[0] = gridValue[1] = valueAtLocation[0] = valueAtLocation[1] = NULL;
      }
      if ((gridPoints=SDDS_RowCount(&SDDSin))<=0)
        SDDS_Bomb("first page of input file is empty");
      if (!(gridValue[0] = SDDS_GetColumnInDoubles(&SDDSin, gridVariable[0])) ||
          !(gridValue[1] = SDDS_GetColumnInDoubles(&SDDSin, gridVariable[1])))
        SDDS_Bomb("grid variables seem to be missing from input file");
      if (!(valueAtLocation[0] = SDDS_GetColumnInDoubles(&SDDSin, findLocationOf[0])) ||
          !(valueAtLocation[1] = SDDS_GetColumnInDoubles(&SDDSin, findLocationOf[1])) )
        SDDS_Bomb("location variables seem to be missing from input file");
      
      gridifyData(gridVariable, gridPoints, presorted);
    }

    if (!findLocationInGrid(atValue[0][iv], atValue[1][iv], &location[0], &value[0], interpolate)) {
      fprintf(stderr, "Couldn't find location for %s=%le, %s=%le\n",
              findLocationOf[0], atValue[0][iv], 
              findLocationOf[1], atValue[1][iv]);
      exit(1);
    } else {
      if (!SDDS_SetRowValues(&SDDSout, SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE, irow++,
                             0, location[0], 1, location[1], 2, value[0], 3, value[1],
                             -1)) {
        SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
      }
    }
    iv++;
  }
  if (!SDDS_WritePage(&SDDSout) || !SDDS_Terminate(&SDDSout) ||
      SDDS_Terminate(&SDDSin))
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
  if (gridValue[0]) {
    free(gridValue[0]);
    free(gridValue[1]);
    free(valueAtLocation[0]);
    free(valueAtLocation[1]);
    gridValue[0] = gridValue[1] = valueAtLocation[0] = valueAtLocation[1] = NULL;
  }
  return (0);
}

double target[2], achieved[2];

double distance(double *position, long *invalid)
{
  uint64_t ix, iy;
  double fx, fy;
  double v[2];
  long i;

  if (position[0]<0 || position[0]>=ng[0] || position[1]<0 || position[1]>=ng[1]) {
    *invalid = 1;
    printf("invalid position: %le, %le for ng=%ld, %ld\n", position[0], position[1],
           (long)ng[0], (long)ng[1]);
    return DBL_MAX;
  }
  *invalid = 0;

  ix = position[0];
  iy = position[1];
  if (ix==(ng[0]-1))
    ix--;
  if (iy==(ng[1]-1))
    iy--;

  fx = position[0] - ix;
  fy = position[1] - iy;

  for (i=0; i<2; i++) {
    double v00, v10, v01, v11, v0, v1;
    v00 = valueAtLocation[i][(ix+0)*ng[1]+(iy+0)];
    v10 = valueAtLocation[i][(ix+1)*ng[1]+(iy+0)];
    v01 = valueAtLocation[i][(ix+0)*ng[1]+(iy+1)];
    v11 = valueAtLocation[i][(ix+1)*ng[1]+(iy+1)];
    v0 = v00 + (v10-v00)*fx;
    v1 = v01 + (v11-v01)*fx;
    v[i] = v0 + (v1-v0)*fy;
    achieved[i] = v[i];
  }

  return sqr(target[0]-v[0]) + sqr(target[1]-v[1]);
}

int findLocationInGrid(double at1, double at2,  double *location, double *value, short interpolate)
{
  uint64_t ix, iy, j;
  uint64_t ixBest, iyBest;
  double delta, bestDelta;
  long i;

  bestDelta = DBL_MAX;
  ixBest = ng[0]/2;
  iyBest = ng[1]/2;
  for (ix=0; ix<ng[0]; ix++) {
    for (iy=0; iy<ng[1]; iy++) {
      j = ix*ng[1] + iy;
      if ((delta = sqr(valueAtLocation[0][j]-at1) + sqr(valueAtLocation[1][j]-at2))<bestDelta) {
        location[0] = gridValue[0][j];
        location[1] = gridValue[1][j];
        value[0] = valueAtLocation[0][j];
        value[1] = valueAtLocation[1][j];
        ixBest = ix;
        iyBest = iy;
        bestDelta = delta;
      }
    }
  }

  if (interpolate) {
    double result, start[2], step[2], lower[2], upper[2];
    start[0] = ixBest;
    start[1] = iyBest;
    step[0] = step[1] = 0.1;
    lower[0] = lower[1] = 0;
    upper[0] = ng[0]-1;
    upper[1] = ng[1]-1;
    target[0] = at1;
    target[1] = at2;
    if (simplexMin(&result, start, step, lower, upper, NULL, 2, 0, 1e-14, distance,
                   NULL, 1500, 3, 12, 3, 1, 0)>=0) {
      double a00, a01, a10, a11, a0, a1, fx, fy;
      ix = start[0];
      iy = start[1];
      if (start[0]<0) ix=0;
      if (ix>=(ng[0]-1)) ix=ng[0]-1;
      if (start[1]<0) iy=0;
      if (iy>=(ng[1]-1)) iy=ng[1]-1;
      fx = start[0] - ix;
      fy = start[1] - iy;

      for (i=0; i<2; i++) {
        a00 = gridValue[i][(ix+0)*ng[1]+(iy+0)];
        a10 = gridValue[i][(ix+1)*ng[1]+(iy+0)];
        a01 = gridValue[i][(ix+0)*ng[1]+(iy+1)];
        a11 = gridValue[i][(ix+1)*ng[1]+(iy+1)];
        a0 = a00 + (a10-a00)*fx;
        a1 = a01 + (a11-a01)*fx;
        location[i] = a0 + (a1-a0)*fy;
        value[i] = achieved[i];
      }
    }
  }

  return 1;
}

int compareGridLocations(const void *data1, const void *data2)
{
  uint64_t i1, i2;
  double diff;
  i1 = *((uint64_t*)data1);
  i2 = *((uint64_t*)data2);
  diff = -(gridValue[0][i1] - gridValue[0][i2]);
  if (diff==0)
    diff = -(gridValue[1][i1] - gridValue[1][i2]);

  return (diff < 0 ? 1 : (diff > 0 ? -1 : 0));
}

void gridifyData(char *gridVariable[2], uint64_t gridPoints, short presorted)
{
  long i;
  char s[256];
  uint64_t j, *index;
  double *buffer;

  for (i=0; i<2; i++) {
    double *copy;
    copy = tmalloc(sizeof(double)*gridPoints);
    memcpy(copy, gridValue[i], gridPoints*sizeof(double));
    qsort((void*)copy, gridPoints, sizeof(double), double_cmpasc);
    ng[i] = 1;
    for (j=1; j<gridPoints; j++) {
      if (copy[j-1]!=copy[j])
        ng[i] += 1;
    }
    free(copy);
    if (ng[i]==gridPoints) {
      snprintf(s, 256, "Grid variable %s has only unique values", gridVariable[i]);
      SDDS_Bomb(s);
    }
    if (ng[i]==1) {
      snprintf(s, 256, "Grid variable %s has only one unique value", gridVariable[i]);
      SDDS_Bomb(s);
    }
  }
  if (ng[0]*ng[1]!=gridPoints) {
    snprintf(s, 256, "input data does not appear to form a grid (nx = %" PRIu64 " , ny=%" PRIu64 ", rows = %" PRIu64")\n",
             ng[0], ng[1], gridPoints);
    SDDS_Bomb(s);
  }
  
  if (!presorted) {
    /* Sort indices for points to place data in gridded order */
    index = tmalloc(sizeof(uint64_t)*gridPoints);
    for (j=0; j<gridPoints; j++)
      index[j] = j;
    
    qsort((void*)index, gridPoints, sizeof(uint64_t), compareGridLocations);
    
    /* copy data into sorted order, in the global variables */
    for (i=0; i<2; i++) {
      buffer = tmalloc(sizeof(double)*gridPoints);
      for (j=0; j<gridPoints; j++)
        buffer[j] = gridValue[i][index[j]];
      free(gridValue[i]);
      gridValue[i] = buffer;
      
      buffer = tmalloc(sizeof(double)*gridPoints);
      for (j=0; j<gridPoints; j++)
        buffer[j] = valueAtLocation[i][index[j]];
      free(valueAtLocation[i]);
      valueAtLocation[i] = buffer;
    }
  }

}


