/*************************************************************************\
 * Copyright (c) 2002 The University of Chicago, as Operator of Argonne
 * National Laboratory.
 * Copyright (c) 2002 The Regents of the University of California, as
 * Operator of Los Alamos National Laboratory.
 * This file is distributed subject to a Software License Agreement found
 * in the file LICENSE that is included with this distribution. 
\*************************************************************************/

/* program: sddsuwrap
 * unwrap data through 
 */
#include "mdb.h"
#include "SDDS.h"
#include "SDDSutils.h"
#include "scan.h"
#include <ctype.h>

static char *USAGE = "sddsunwrap [<input>] [<output>]\n\
[-pipe=[input][,output]] [-column=list of columns] [-threshold=<value>] [-modulo=<value>] \n\
-threshold    discontinuity threshold used to identify a wrap in the data, default is PI.\n\
-modulo       the value used to unwrap the data, default is 2*PI.\n\
-column       list of column names to be unwrapped, separated by comma, accept wild cards\n\n\
sddsunwrap    looks for discontinuities of size > threshold in a set of data.  \n\
              After each discontinuity it adds the appropriate multiple of the modulo to the data set. .\n\
Program by Hairong Shang.  (" __DATE__ " " __TIME__ ", SVN revision: " SVN_VERSION ")";

#define CLO_PIPE 0
#define CLO_COLUMN 1
#define CLO_THRESHOLD 2
#define CLO_MAJOR_ORDER 3
#define CLO_MODULO 4
#define CLO_OPTIONS 5
static char *option[CLO_OPTIONS] = {
  "pipe",
  "column",
  "threshold",
  "majorOrder",
  "modulo",
};

int main(int argc, char **argv) {
  double *xData, *yData, threshold, modulo = 0;
  char *input = NULL, *output = NULL, **column_match = NULL, **columnName = NULL, **column;
  long i, iArg;
  char outputColumn[256];

  SDDS_DATASET SDDSin, SDDSout;
  SCANNED_ARG *scanned;
  unsigned long pipeFlags = 0, majorOrderFlag = 0;
  int32_t columnType, columns0 = 0;
  int64_t j, k, rows;
  int32_t columns = 0, column_matches = 0;
  short columnMajorOrder = -1, phase = 1, thresholdProvided = 0;

  SDDS_RegisterProgramName(argv[0]);

  argc = scanargs(&scanned, argc, argv);
  if (argc == 1)
    bomb(NULL, USAGE);

  input = output = NULL;
  xData = yData = NULL;
  columnName = NULL;

  for (iArg = 1; iArg < argc; iArg++) {
    if (scanned[iArg].arg_type == OPTION) {
      /* process options here */
      switch (match_string(scanned[iArg].list[0], option, CLO_OPTIONS, 0)) {
      case CLO_MAJOR_ORDER:
        majorOrderFlag = 0;
        scanned[iArg].n_items--;
        if (scanned[iArg].n_items > 0 &&
            (!scanItemList(&majorOrderFlag, scanned[iArg].list + 1, &scanned[iArg].n_items, 0,
                           "row", -1, NULL, 0, SDDS_ROW_MAJOR_ORDER,
                           "column", -1, NULL, 0, SDDS_COLUMN_MAJOR_ORDER, NULL)))
          SDDS_Bomb("invalid -majorOrder syntax/values");
        if (majorOrderFlag & SDDS_COLUMN_MAJOR_ORDER)
          columnMajorOrder = 1;
        else if (majorOrderFlag & SDDS_ROW_MAJOR_ORDER)
          columnMajorOrder = 0;
        break;
      case CLO_THRESHOLD:
        if (scanned[iArg].n_items != 2)
          SDDS_Bomb("invalid -threshold syntax");
        if (!get_double(&threshold, scanned[iArg].list[1]))
          SDDS_Bomb("invalid -threshold value given");
        thresholdProvided = 1;
        break;
      case CLO_MODULO:
        if (scanned[iArg].n_items != 2)
          SDDS_Bomb("invalid -modulo syntax");
        if (!get_double(&modulo, scanned[iArg].list[1]))
          SDDS_Bomb("invalid -modulo value given");
        break;
      case CLO_COLUMN:
        if ((scanned[iArg].n_items < 2))
          SDDS_Bomb("invalid -column syntax");
        column_matches = scanned[iArg].n_items - 1;
        column_match = tmalloc(sizeof(*column_match) * column_matches);
        for (i = 0; i < column_matches; i++)
          column_match[i] = scanned[iArg].list[i + 1];
        break;
      case CLO_PIPE:
        if (!processPipeOption(scanned[iArg].list + 1, scanned[iArg].n_items - 1, &pipeFlags))
          SDDS_Bomb("invalid -pipe syntax");
        break;
      default:
        fprintf(stderr, "Unknown option %s provided\n", scanned[iArg].list[0]);
        exit(1);
        break;
      }
    } else {
      if (!input)
        input = scanned[iArg].list[0];
      else if (!output)
        output = scanned[iArg].list[0];
      else
        SDDS_Bomb("too many filenames");
    }
  }
  processFilenames("sddsunwrap", &input, &output, pipeFlags, 0, NULL);
  if (phase && !thresholdProvided)
    threshold = PI;

  if (!modulo)
    modulo = 2 * threshold;

  if (!SDDS_InitializeInput(&SDDSin, input))
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
  if (column_matches) {
    column = getMatchingSDDSNames(&SDDSin, column_match, column_matches, &columns, SDDS_MATCH_COLUMN);
  } else {
    column = tmalloc(sizeof(*column) * 1);
    if (!(columnName = (char **)SDDS_GetColumnNames(&SDDSin, &columns0)))
      SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
    for (i = 0; i < columns0; i++) {
      if (SDDS_NUMERIC_TYPE(columnType = SDDS_GetColumnType(&SDDSin, i))) {
        SDDS_CopyString(&column[0], columnName[i]);
        break;
      }
    }
    SDDS_FreeStringArray(columnName, columns0);
    columns = 1;
  }

  if (!SDDS_InitializeCopy(&SDDSout, &SDDSin, output, "w"))
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
  if (columnMajorOrder != -1)
    SDDSout.layout.data_mode.column_major = columnMajorOrder;
  else
    SDDSout.layout.data_mode.column_major = SDDSin.layout.data_mode.column_major;

  for (i = 0; i < columns; i++) {
    sprintf(outputColumn, "Unwrap%s", column[i]);
    if (!SDDS_TransferColumnDefinition(&SDDSout, &SDDSin, column[i], outputColumn))
      SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
  }
  if (!SDDS_WriteLayout(&SDDSout))
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);

  while (SDDS_ReadPage(&SDDSin) > 0) {
    if ((rows = SDDS_CountRowsOfInterest(&SDDSin)) <= 0)
      continue;
    if (!SDDS_StartPage(&SDDSout, rows) || !SDDS_CopyPage(&SDDSout, &SDDSin))
      SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);

    yData = tmalloc(sizeof(*yData) * rows);
    for (i = 0; i < columns; i++) {
      sprintf(outputColumn, "Unwrap%s", column[i]);
      if (!(xData = SDDS_GetColumnInDoubles(&SDDSin, column[i])))
        SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
      k = 0;
      for (j = 0; j < rows - 1; j++) {
        yData[j] = xData[j] + modulo * k; /*add 2*Pi*k to input */
        /* check if diff is greater than alpha */
        if (fabs(xData[j + 1] - xData[j]) > fabs(threshold)) {
          if (xData[j + 1] < xData[j])
            k = k + 1; /* if the pahse jump is negative, increment k */
          else
            k = k - 1; /* if the phase jump is positive, decrement k */
        }
      }
      yData[rows - 1] = xData[rows - 1] + modulo * k; /*  add 2*pi*k to the last element of the input */

      if (xData)
        free(xData);
      xData = NULL;

      if (!SDDS_SetColumnFromDoubles(&SDDSout, SDDS_BY_NAME, yData, rows, outputColumn))
        SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
    }
    free(yData);
    yData = NULL;
    if (!SDDS_WritePage(&SDDSout))
      SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
  }
  if (!SDDS_Terminate(&SDDSin) || !SDDS_Terminate(&SDDSout)) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors);
    exit(1);
  }
  if (columnName) {
    SDDS_FreeStringArray(columnName, columns0);
    free(columnName);
  }
  SDDS_FreeStringArray(column, columns);
  free(column);
  if (column_match)
    free(column_match);
  free_scanargs(&scanned, argc);
  return (0);
}
